/*
   Module .............libbt
   File ...............SimpleControl.c
   Author .............Traveler Hauptman
   Creation Date ......Nov 24, 2002
/*======================================================================*
 *  Module .............libbtsystem
 *  File ...............SimpleControl.c
 *  Author .............Traveler Hauptman
 *  Creation Date ......Nov 24, 2002
 *  Addtl Authors ......Brian Zenowich, Sam Clanton
 *                                                                      *
 *  ******************************************************************  *
 *
 *  NOTES:
 *   A simple Torque, PID & Trapezoidal velocity trajectory controller
 *
 *  REVISION HISTORY:
 *  021124 - TH - File created
 *  16 Dec 2004 - BZ, SC, TH
 *    Initial port to linux + RTAI
 *
 *======================================================================*/

/*! \file SimpleControl.c
    \brief A simple Torque, PID & Trapezoidal velocity trajectory controller

    SimpleControl provides a set of basic control algorithms that are often useful
    when controlling robots.
    The use of these functions is straighforward
    -# Create a sturcture of type SimpleCtl
    -# SCinit to get everything in a reasonable state.
    -# SCevaluate inside your read-position set-torque loop.
    -# SCsetmode to switch to the mode you want

    - SCstarttrj to move the PID setpoint along a trapezoidal trajectory

*/
#include <math.h>

//th041216#include <sys/neutrino.h>

#include <pthread.h>
#include <errno.h>
#include <syslog.h>
#include "SimpleControl.h"
#include "btcan.h"

#define sign(x) (x>=0?1:-1)
#define Sgn(x) (x>=0.0?1.0:-1.0)

/*! Initialize the SimpleCtl structure */
int SCinit(SimpleCtl *sc)
{
    int err;

    sc->mode = MODE_IDLE;
    sc->requested_torque = 0.0;
    sc->kill_torque = 500000;
    sc->last_dt = 1;
    sc->command_torque = 0;
    sc->trj.state = 0;
    sc->trj.acc = .00000001;
    sc->trj.vel = .00000001;
    sc->via.state = 0;
    PIDinit(&(sc->pid),0, 0, 0, 0);
    err = pthread_mutex_init(&(sc->mutex),NULL);
    if (err)
    {
        syslog(LOG_ERR, "Could not initialize  mutex for simple controller.");
    }
}

/*! Evaluate the Simple Controller

  See the documentation for SimpleCtl for an overview.

  This function takes as input the current position and the time since the last evaluation.
It returns a torque.

\param position the measured position
\param dt the time step we want to evaluate with.
\return Returns a torque command.
*/
double SCevaluate(SimpleCtl *sc, double position, double dt)
{
    double newcommand;
    double newtorque;
    double cmptorque;
    int err;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCevaluate lock mutex failed: %d", err);
    }

    sc->position = position;
    switch (sc->mode)
    {
        case MODE_IDLE://Idle
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != 0)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (idle) failed: %d", err);
            }
            sc->command_torque = 0.0;
            return sc->command_torque;
            break;
        case MODE_TORQUE://Torque
            sc->command_torque = sc->requested_torque;
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != 0)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (Torque) failed: %d", err);
            }
            return sc->command_torque;
            break;
        case MODE_PID://PID

            sc->pid.y = sc->position;
            sc->pid.dt = dt;
            newtorque = PIDcalc(&(sc->pid));
            cmptorque = fabs(newtorque);

            if (cmptorque > sc->kill_torque)
            {
                syslog(LOG_ERR, "sc->kill_torque: pid.y=%lf pid.yref=%lf trq=%lf", sc->pid.y, sc->pid.yref, newtorque);
                err = pthread_mutex_unlock( &(sc->mutex) );
                if(err != 0)
                {
                    syslog(LOG_ERR, "SCevaluate:kill_torque unlock mutex (PID) failed: %d", err);
                }
                SCsetmode(sc,MODE_PID);
                err = pthread_mutex_lock( &(sc->mutex) );
                if(err != 0)
                {
                    syslog(LOG_ERR, "SCevaluate:kill_torque lock mutex failed: %d", err);
                }
                newtorque = 0.0;
            }
            sc->command_torque = newtorque;
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != 0)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (PID) failed: %d", err);
            }
            return newtorque;
            break;
        default:
            sc->error = 1;
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != 0)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (default) failed: %d", err);
            }
            return 0.0;
        }

}



/*! \brief Set the controller mode

  Sets the controller mode. To keep anything from "jumping" unexpectedly, we reset our
  pid controller whenever we switch into pidmode.

*/
int SCsetmode(SimpleCtl *sc, int mode)
{
    int err;
    double tmp;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsetmode lock mutex failed: %d", err);
    }
    if (sc->trj.state != TRAJ_DONE) sc->trj.state = TRAJ_DONE;
    sc->pid.y = sc->position;
    PIDreset(&(sc->pid));
    tmp = PIDcalc(&(sc->pid));
    sc->mode = mode;
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsetmode unlock mutex failed: %d", err);
    }
    return 0;
}


int SCsettorque(SimpleCtl *sc,double torque)
{
    int err;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsettorque lock mutex failed: %d", err);
    }
    sc->requested_torque = torque; //use a buffer since we are not necessarily running from the same thread as evaluatecontroller
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsettorque unlock mutex failed: %d", err);
    }

    if (sc->mode != MODE_TORQUE)
    {
        return -1;
    }
    return 0;
}

int SCsetpid(SimpleCtl *sc,double kp,double kd,double ki, double saturation) //!! not thread safe yet! We may update these in the middle of their use.
{

    int err;

    if (sc->trj.state == TRAJ_RUN || sc->mode == MODE_PID) //make sure we don't screw with stuff in the middle of a trajectory.
        return -1;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsetpid lock mutex failed: %d", err);
    }
    sc->pid.Kp = kp;
    sc->pid.Kd = kd;
    sc->pid.Ki = ki;
    sc->pid.saturation = saturation;

    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsetpid unlock mutex failed: %d", err);
    }

    return 0;
}

int SCsetcmd(SimpleCtl *sc,double cmd) //sets pid command position. if in traj mode, stops trajectory
{
    int err;

    //if (sc->trj.state != TRAJ_DONE)
    //    return -1;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsetcmd lock mutex failed: %d", err);
    }
    sc->pid.yref = cmd;
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsetcmd unlock mutex failed: %d", err);
    }
    return 0;
}

int SCsettrjprof(SimpleCtl *sc,double vel,double acc)
{
    int err;

    if (sc->trj.state == TRAJ_RUN)
        return -1;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsettrjprof lock mutex failed: %d", err);
    }
    sc->trj.vel = fabs(vel);
    sc->trj.acc = fabs(acc);

    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCsettrjprof unlock mutex failed: %d", err);
    }
    return 0;

}
int SCstarttrj(SimpleCtl *sc,double end)
{
    int err;
    double here;

    if (sc->trj.state != TRAJ_DONE)
        return -1;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCstarttrj lock mutex failed: %d", err);
    }
    here = sc->pid.yref;
    init_trajectory(&(sc->trj),here,end);

    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCstarttrj unlock mutex failed: %d", err);
    }
    return 0;
}

int SCstartTrjFile(SimpleCtl *sc, double ***trajFile, int rows)
{
    int err;

    if (sc->trj.state != TRAJ_DONE)
        return -1;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCstartTrjFile lock mutex failed: %d", err);
    }
    sc->trj.t = 0;
    sc->trj.rows = rows;
    sc->trj.tf = *trajFile;
    sc->trj.state = TRAJ_FILE;

    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCstartTrjFile unlock mutex failed: %d", err);
    }
    return 0;
}
int SCstartViaTrjFile(SimpleCtl *sc, double ***trajFile, int n,int col,double acc)
{
    int err;
    double here;

    if (sc->trj.state != TRAJ_DONE)
        return -1;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCstartTrjFile lock mutex failed: %d", err);
    }

    here = sc->pid.yref;
    sc->via.tf = *trajFile;
    if (n < 3) syslog(LOG_ERR, "SCstartTrjFile not enough points");
    else start_via_trj(&(sc->via),here,n,col+1,acc);

    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != 0)
    {
        syslog(LOG_ERR, "SCstartTrjFile unlock mutex failed: %d", err);
    }
    return 0;
}

/*! Calculates all the variables necessary to set up a trajectory.

\param start the starting position
\param end the destination position


*/
void init_trajectory(trajectory *traj, double start, double end) //assumes that velocity and acceleration have been set to reasonable values
{
    double ax; //acceleration x
    double dist;


    traj->cmd = 0;
    traj->start_point = start;
    traj->end_point = end;

    traj->t1 = (traj->vel / traj->acc); //Calculate the "ramp-up" time

    dist = traj->end_point - traj->start_point;

    if (dist > 0) traj->sign = 1;
    else traj->sign = -1;

    //Get the point at which acceleration stops
    ax = 0.5*traj->acc*traj->t1*traj->t1; //x = 1/2 a t^2

    dist = fabs(dist);
    traj->end = dist;
    if (ax > dist/2) //If the acceleration completion point is beyond the halfway point
        ax = dist/2; //Stop accelerating at the halfway point
    traj->x1 = ax;   //Set the top left point of the trapezoid
    traj->x2 = dist - ax; //Set the top right point of the trapezoid
    traj->t2 = (dist-(2*ax)) / traj->vel + traj->t1; //Find the time to start the "ramp-down"
    traj->t = 0;
    traj->state = TRAJ_RUN;
    if (dist == 0.0) traj->state = TRAJ_DONE;
}

double calc_traj_time(trajectory *traj, double start, double end, double vel, double acc)
{
    double ax; //acceleration x
    double dist,t1,t2,time;


    t1 = (vel / acc); //Calculate the "ramp-up" time

    dist = end - start;

    //Get the point at which acceleration stops
    ax = 0.5*acc*t1*t1; //x = 1/2 a t^2
    dist = fabs(dist);

    if (ax > dist/2) //If we never reach max velacity, back calculate for time
    {
        ax = dist/2; //Stop accelerating at the halfway point
        time = 2*sqrt(2*ax/acc);
    }
    else
    {
        time = (dist-(2*ax)) / vel + 2*t1;
    }
    return time;
}

void sync_trajectory(trajectory *traj, double start, double end, double vel, double acc, double t)
{
    double ax; //acceleration x
    double dist,t1,t2,time;


    t1 = (vel / acc); //Calculate the "ramp-up" time
    if (2*t1 > t) t1 = t;
    t2 = t - 2*t1;
    dist = fabs(end - start);

    traj->acc= dist/(t1*t1+t1*t2);
    traj->vel= traj->acc*t1;
}

double scale_vel(trajectory *traj, double start, double end, double vel, double acc, double t)
{
    double ax; //acceleration x
    double dist,t1,t2,time,acc1;


    t1 = (vel / acc); //Calculate the "ramp-up" time
    if (2*t1 > t) t1 = t;
    t2 = t - 2*t1;
    dist = fabs(end - start);

    acc1= dist/(t1*t1+t1*t2);
    return (acc1*t1);
}
double scale_acc(trajectory *traj, double start, double end, double vel, double acc, double t)
{
    double ax; //acceleration x
    double dist,t1,t2,time,acc1;


    t1 = (vel / acc); //Calculate the "ramp-up" time
    if (2*t1 > t) t1 = t;
    t2 = t - 2*t1;
    dist = fabs(end - start);
    acc1 = dist/(t1*t1+t1*t2);
    return(acc1);
}

/*! Calculate next position on the trajectory

evaluate_trajectory generates points on a trapezoidal velocity trajectory. The points accerate with
constant acceleration specified by acc up to a maximum velocity specified by vel. The max velocity
is maintained until approximately the time it will take to decellerate at acc to a velocity of zero.

\param *traj pointer to the trajectory structure
\param dt the time step to take (in the same units as vel and acc)

\return The value returned is the next point on the trajectory after a step of dt.


*/
double evaluate_trajectory(trajectory *traj,double dt)
{
    double remaining_time,newtime,result;

    if (traj->state == TRAJ_RUN)
    {
        traj->t += dt;
        if (traj->cmd < traj->x1) //If we are in "accel" stage
            traj->cmd = 0.5 * traj->acc * traj->t * traj->t;  //x = 1/2 a t^2
        else if (traj->cmd < traj->x2) //If we are in "cruise" stage
            traj->cmd = traj->x1 + traj->vel * (traj->t - traj->t1); //x = x + v t
        else //We are in "decel" stage
        {
            /* time to hit zero = sqrt( (target position - current position)*2/a)
                                 new position = 1/2 acc * (time to hit zero - dt)^2 */
            remaining_time = sqrt((traj->end - traj->cmd)*2/traj->acc);
            if (dt > remaining_time)
            {
                traj->cmd = traj->end;
                traj->state = TRAJ_DONE;
            }

            newtime = (remaining_time-dt);
            if(newtime <= 0)
            {
                traj->cmd = traj->end;
                traj->state = TRAJ_DONE;
            }
            traj->cmd =  traj->end - 0.5 * traj->acc * newtime * newtime;
        }

    }
    result = (traj->start_point + traj->sign*traj->cmd);

    if (isnan(result))
    {
        syslog(LOG_ERR, "nan in eval_traj");
        traj->cmd = traj->end;
        traj->state = TRAJ_DONE;
        return traj->end_point;
    }
    return (traj->start_point + traj->sign*traj->cmd);
}


/*! Initializes the PIDregulator data structure

The function sets the main parameters of a PID regulator. You
will only have to use this function a single time when you create the variable. Before
you start using PIDcalc, remember to set y and yref.
*/
void PIDinit(PIDregulator *pid,double Kp, double Kd, double Ki, double dt)
{
    pid->Kp = Kp;
    pid->Kd = Kd;
    pid->Ki = Ki;
    pid->dt = dt;
    pid->filter_de = 0;
    pid->se = 0;
    pid->e = 0;
    pid->laste = 0;
    pid->de = 0;
    pid->firsttick = 1;
    pid->y = 0;
    pid->yref = 0;
    pid->lastresult = 0;
    pid->de_filter.order = 2;
    pid->de_filter.cutoffHz = 10;
    InitDifferentiator (&(pid->de_filter), .001);
    pid->se_filter.order = 2;
    pid->se_filter.cutoffHz = 10;
    InitDifferentiator (&(pid->se_filter), .001);
    //syslog(LOG_ERR,"%f,%f",pid->de_filter.d[0],pid->de_filter.d[1]);


}
/** Reset the state parameters of the PID regulator

  Sets the accumulated error to zero, the reference value to the present value, and notifies
  the algorithm that it is re-starting.
*/
void PIDreset(PIDregulator *pid)
{
    pid->se = 0;
    pid->firsttick = 1;
    pid->yref = pid->y;
}

/** Increments the state by dt and returns the output of the regulator.

PIDcalc updates the PIDregulator state variable by incrementing the time by dt
and then calculates the error between y and yref.

\param *pid A pointer to the pre-allocated and initialized PIDregulator structure.

\return The regulator output for the present dt, y, and yref.
*/
double PIDcalc(PIDregulator *pid)
{
    pid->e = pid->yref - pid->y;

    if (pid->firsttick)
    {
        pid->laste = pid->e;
        pid->firsttick = 0;
    }

    pid->de = (pid->e-pid->laste)/pid->dt;  //Backward euler
    if (pid->filter_de)
        pid->de = Filter(&(pid->de_filter),pid->e);
    pid->laste = pid->e;

    pid->se += pid->e;
    pid->fe = Filter(&(pid->se_filter),pid->se);
    if (pid->saturation != 0)
        if ((fabs(pid->se)*pid->Ki) > pid->saturation)
        pid->se = sign(pid->se)*pid->saturation;

    pid->lastresult = pid->Kp*pid->e+pid->Kd*pid->de+pid->Ki*pid->se;

    return(pid->lastresult); //bz
}

/*****************************************************************/
/* filtered_value = F i l t e r ( unfiltered_value ) */
/*****************************************************************/
double Filter (filter_struct *filt, double xnew)
{
    int       i,idx;
    double      ynew;

    idx = filt->index;

    filt->x[idx] = xnew;
    filt->y[idx] = 0;

    for (i=0; i<=filt->order; i++)
        filt->y[idx] += filt->n[i] * filt->x[ (idx+i+1)%(filt->order+1) ]
    - filt->d[i] * filt->y[ (idx+i+1)%(filt->order+1) ];

    ynew  = filt->y[idx];
    filt->index = (++idx) % (filt->order+1);

    return(ynew);
}

/** A filter that doesn't do anything
  This function creates a filter that multiplies the input by 1
**/
void InitPassthroughFilter (filter_struct *filt)
{
    int i;

    filt->index = 0;
    for(i=0; i<=4; i++)
    {
        filt->x[i] = 0;
        filt->y[i] = 0;
        filt->d[i] = 0;
        filt->n[i] = 0;
    }

    filt->n[0] = 1;

}

/************************************************************/
/*   I n i t D i f f e r e n t i a t o r ( )                    */
/************************************************************/
void InitDifferentiator (filter_struct *filt, double sample_time)
{
    //  Initializes coefficients for differentiator plus first or
    //  second order lowpass filter.

    // You must initialize the following arguments in main:
    //    diff.order, diff.cutoffHz, sample_time

    double woT;
    double temp;
    int i;

    filt->index = 0;
    for(i=0; i<=4; i++)
    {
        filt->x[i] = 0;
        filt->y[i] = 0;
    }

    if(filt->order == 1)
    {
        woT = sample_time*filt->cutoffHz*2.0*PI;
        temp = 1.0/(2+woT);

        filt->d[1] = 0.0;
        filt->d[0] = -(2-woT)*temp;

        filt->n[1] = 2.0*filt->cutoffHz*PI*2.0*temp;
        filt->n[0] = -(filt->n[1]);
    }
    else if(filt->order == 2)
    {
        woT = sample_time*filt->cutoffHz*2.0*PI;
        temp = 1.0/ ( 4.0*(1.0+woT) + woT*woT  );

        filt->d[2] = 0.0;
        filt->d[1] = -(8.0 - 2.0*woT*woT) * temp;
        filt->d[0] = ( 4.0*(1.0-woT) + woT*woT  ) * temp;

        filt->n[2] = 2.0*filt->cutoffHz*PI*2.0*woT * temp;
        filt->n[1] = 0.0;
        filt->n[0] = -(filt->n[2]);
    }
    else
    {
        syslog(LOG_ERR,"Filter order must be 1 or 2.\n");

    }

}
void CalcSegment(Seg_int *seg,double q1, double q2, double t1, double t2, double v_prev, double v_next, double seg_acc, int end)
{
    double dt,dq;
    double vel,acc1,acc2,dt_vel,dt_acc1,dt_acc2;

           
    dt = t2 - t1;
    dq = q2 - q1;

    if (end == 0)
    {
        acc1 = Sgn(dq)*seg_acc;
        dt_acc1 = dt - sqrt(dt*dt - 2*dq/acc1);
        vel = dq/(dt - dt_acc1);
        acc2 = Sgn(v_next - vel)*seg_acc;
        dt_acc2 = (v_next - vel)/acc2;
        dt_vel = dt - dt_acc1 - dt_acc2/2;
    }
    else if (end == 1)
    {
        vel = dq/dt;
        acc1 = Sgn(vel - v_prev)*seg_acc;
        dt_acc1 = (vel - v_prev)/acc1;
        acc2 = Sgn(v_next - vel)*seg_acc;
        dt_acc2 = (v_next - vel)/acc2;
        dt_vel = dt - dt_acc1/2 - dt_acc2/2;
    }
    else
    {
        acc2 = Sgn(dq)*seg_acc;
        dt_acc1 = dt - sqrt(dt*dt + 2*dq/acc2);
        vel = dq/(dt - dt_acc2);
        acc1 = Sgn((vel - v_prev))*seg_acc;
        dt_acc1 = ((vel - v_prev))/acc1;
        dt_vel = dt - dt_acc2 - dt_acc1/2;
    }

    seg->vel = vel;
    seg->acc1 = acc1;
    seg->acc2 = acc2;
    seg->dt_vel = dt_vel;
    seg->dt_acc1 = dt_acc1;
    seg->dt_acc2 = dt_acc2;
}

void start_via_trj(via_trj *trj,double start,int n,int col,double acc)
{
    double Dt,Dq,Dv,Dt_next,Dq_next,Dv_next;

    trj->idx = 0;
    trj->n = n;
    trj->col = col;
    (trj->tf)[trj->idx][trj->col] = start;
    trj->trj_acc = acc;

    trj->segment = 0;//acc first

    Dt_next = (trj->tf)[trj->idx+2][0] - (trj->tf)[trj->idx+1][0];
    Dq_next = (trj->tf)[trj->idx+2][trj->col] - (trj->tf)[trj->idx+1][trj->col];
    trj->v_prev = 0;
    trj->v_next = Dq_next/Dt_next;
    CalcSegment(&(trj->seg),(trj->tf)[trj->idx][trj->col], (trj->tf)[trj->idx+1][trj->col],
                (trj->tf)[trj->idx][0], (trj->tf)[trj->idx+1][0],trj->v_prev,trj->v_next,trj->trj_acc, 0);

    trj->acc = trj->seg.acc1;
    trj->dt_acc = trj->seg.dt_acc1;
    trj->t=0;
    trj->t0=0;
    trj->t_acc = trj->dt_acc;
    trj->q0 = (trj->tf)[trj->idx][trj->col];
    trj->v0 = 0;
    trj->state = TRAJ_RUN;

    /*syslog(LOG_ERR, "col=%d idx=%d vel=%f t_vel=%f t_acc=%f q0=%f v0=%f t=%f t0=%f q1=%f q2=%f t1=%f t2=%f",
           trj->col,trj->idx,trj->vel,trj->t_vel,trj->t_acc,
           trj->q0,trj->v0,trj->t,trj->t0,(trj->tf)[trj->idx][trj->col], (trj->tf)[trj->idx+1][trj->col],
                (trj->tf)[trj->idx][0], (trj->tf)[trj->idx+1][0]); */
}

double eval_via_trj(via_trj *trj,double dt)
{
    double Dt,Dq,Dv,Dt_next,Dq_next,Dv_next,acc_next,t_acc_next,end,et,cmd;
    if (trj->state == TRAJ_RUN)
    {
        trj->t += dt;

        if ((trj->segment == 0) && (trj->t > trj->t_acc))
        { //done with acc, set up vel
            if (trj->idx >= trj->n-1)
            {
                trj->state = TRAJ_DONE;
            }
            else
            {
                trj->vel = trj->seg.vel;
                trj->dt_vel = trj->seg.dt_vel;
                trj->t0 = trj->t_acc;
                trj->t_vel = trj->dt_vel + trj->t0;
                trj->q0 = trj->q0 + trj->v0*trj->dt_acc + 0.5*trj->dt_acc*trj->dt_acc*trj->acc;
                trj->segment = 1;
            }

            /*syslog(LOG_ERR, "col=%d idx=%d vel=%f t_vel=%f t_acc=%f q0=%f v0=%f t=%f t0=%f q1=%f q2=%f t1=%f t2=%f",
           trj->col,trj->idx,trj->vel,trj->t_vel,trj->t_acc,
           trj->q0,trj->v0,trj->t,trj->t0,(trj->tf)[trj->idx][trj->col], (trj->tf)[trj->idx+1][trj->col],
                (trj->tf)[trj->idx][0], (trj->tf)[trj->idx+1][0]); */
        }
        else if((trj->segment == 1) && (trj->t > trj->t_vel))
        {//setup acc segment
            trj->idx++;
            if (trj->idx >= trj->n-1)
            {
                trj->acc = trj->seg.acc2;
                trj->dt_acc = trj->seg.dt_acc2;
                trj->t0 = trj->t_vel;
                trj->t_acc = trj->dt_acc + trj->t0;
                trj->q0 = trj->q0 + trj->vel*trj->dt_vel;
                trj->v0 = trj->vel;
                trj->segment = 0;//acc first
            }
            else
            {
              if (trj->idx >= trj->n-2){
                end = 2;
                Dt = (trj->tf)[trj->idx][0] - (trj->tf)[trj->idx-1][0];
                Dq = (trj->tf)[trj->idx][trj->col] - (trj->tf)[trj->idx-1][trj->col];
                trj->v_prev = Dq/Dt;
                trj->v_next = 0;
         
              }
              else{
                    end = 1;

                Dt = (trj->tf)[trj->idx][0] - (trj->tf)[trj->idx-1][0];
                Dq = (trj->tf)[trj->idx][trj->col] - (trj->tf)[trj->idx-1][trj->col];
                Dt_next = (trj->tf)[trj->idx+2][0] - (trj->tf)[trj->idx+1][0];
                Dq_next = (trj->tf)[trj->idx+2][trj->col] - (trj->tf)[trj->idx+1][trj->col];
                trj->v_prev = Dq/Dt;
                trj->v_next = Dq_next/Dt_next;
              }
              CalcSegment(&(trj->seg),(trj->tf)[trj->idx][trj->col], (trj->tf)[trj->idx+1][trj->col],
                            (trj->tf)[trj->idx][0], (trj->tf)[trj->idx+1][0],trj->v_prev,trj->v_next,trj->trj_acc, end);
                trj->segment = 0;//acc first
                trj->acc = trj->seg.acc1;
                trj->dt_acc = trj->seg.dt_acc1;
                trj->t0 = trj->t_vel;
                trj->t_acc = trj->dt_acc + trj->t0;
                trj->q0 = trj->q0 + trj->vel*trj->dt_vel;
                trj->v0 = trj->vel;
            }

           /*syslog(LOG_ERR, "col=%d idx=%d vel=%f t_vel=%f t_acc=%f q0=%f v0=%f t=%f t0=%f q1=%f q2=%f t1=%f t2=%f",
           trj->col,trj->idx,trj->vel,trj->t_vel,trj->t_acc,
           trj->q0,trj->v0,trj->t,trj->t0,(trj->tf)[trj->idx][trj->col], (trj->tf)[trj->idx+1][trj->col],
                (trj->tf)[trj->idx][0], (trj->tf)[trj->idx+1][0]); */
        }



        if (trj->segment == 0)
        {//in acceleration segment
            et = trj->t - trj->t0;
            cmd = trj->q0 + trj->v0*et + 0.5*et*et*trj->acc;
        }
        else
        {
            et = trj->t - trj->t0; //elapsed time
            cmd = trj->q0 + trj->vel*et;
        }

    }
    //syslog(LOG_ERR, "VIA %d %d %f", trj->idx, trj->col, cmd);
    return cmd;

}




double linterp(double x1, double y1, double x2, double y2, double x)
{
    if ((x2-x1) != 0.0)
        return (x*(y2-y1)/(x2-x1));
    else
        return (y2+y1)/2;
}

/*======================================================================*
 *                                                                      *
 *          Copyright (c) 2003, 2004 Barrett Technology, Inc.           *
 *                        625 Mount Auburn St                           *
 *                    Cambridge, MA  02138,  USA                        *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *======================================================================*/


